/*
 * Copyright (C) 2009 Igalia S.L.
 *
 * Contact: Iago Toral Quiroga <itoral@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include "mafw-fs-source.h"

#include <libgnomevfs/gnome-vfs.h>
#include <gmodule.h>
#include <string.h>

struct browse_closure {
	MafwSource *source;
	gchar *objectid;
	gchar **keys;
	guint offset;
	guint count;
	MafwSourceBrowseResultCb browse_cb;
	gpointer user_data;
	guint browseid;
};

struct metadata_closure {
	MafwSource *source;
	gchar *objectid;
	gchar **keys;
	MafwSourceMetadataResultCb metadata_cb;
	gpointer user_data;
};

struct metadatas_closure {
	MafwSource *source;
	gchar **objectids;
	gchar **keys;
	MafwSourceMetadataResultsCb metadata_cb;
	gpointer user_data;
};

#define MAFW_FS_SOURCE_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE((object), MAFW_TYPE_FS_SOURCE, \
                                      MafwFsSourcePrivate))

static void mafw_fs_source_get_metadatas (MafwSource *source,
					  const gchar **objectids,
					  const gchar *const *keys,
					  MafwSourceMetadataResultsCb metadata_cb,
					  gpointer user_data);

static void mafw_fs_source_get_metadata (MafwSource *source,
					 const gchar *objectid,
					 const gchar *const *keys,
					 MafwSourceMetadataResultCb metadata_cb,
					 gpointer user_data);

static guint mafw_fs_source_browse (MafwSource *source,
				    const gchar *objectid,
				    gboolean recursive,
				    const MafwFilter *filter,
				    const gchar *sorting,
				    const gchar *const *keys,
				    guint offset,
				    guint count,
				    MafwSourceBrowseResultCb browse_cb,
				    gpointer user_data);

struct _MafwFsSourcePrivate {
	guint browseid;
};

/*________________________ Plugin init  ________________________*/

/*
 * Plugin descriptor
 */
G_MODULE_EXPORT MafwPluginDescriptor mafw_fs_source_plugin_description =
{
	{ .name = MAFW_FS_SOURCE_PLUGIN_NAME },
	.initialize = mafw_fs_source_plugin_initialize,
	.deinitialize = mafw_fs_source_plugin_deinitialize,
};

/*
 * Plugin initialization. Creates an instance of the source and 
 * registers it so it becomes available for applications. A plugin
 * could spawn more than one source instance, and each one must have
 * a unique UUID (see mafw-upnp-source for an example of this).
 */
gboolean
mafw_fs_source_plugin_initialize (MafwRegistry *registry, GError **error)
{
	MafwSource *source;

	/* Start Gnome VFS */
	gnome_vfs_init ();

	/* Create a filesystem source instance and register it */
	source = MAFW_SOURCE (
		g_object_new (MAFW_TYPE_FS_SOURCE,
			      "plugin", MAFW_FS_SOURCE_PLUGIN_NAME,
			      "uuid", MAFW_FS_SOURCE_UUID,
			      "name", MAFW_FS_SOURCE_NAME,
			      NULL));

	mafw_registry_add_extension (registry, MAFW_EXTENSION (source));
	
	return TRUE;
}

/*
 * Plugin deinit, here we should free any resources used by the plugin.
 */
void 
mafw_fs_source_plugin_deinitialize (GError **error) 
{
	/* Shut down Gnome VFS */
	gnome_vfs_shutdown ();
}


/*_____________________ GObject setup ________________________ */


G_DEFINE_TYPE (MafwFsSource, mafw_fs_source, MAFW_TYPE_SOURCE);

static void
mafw_fs_source_class_init (MafwFsSourceClass *klass)
{
	MafwSourceClass *source_class = MAFW_SOURCE_CLASS (klass);

	/* We only implement a subset of the MafwSource's API */
	source_class->browse = mafw_fs_source_browse;
	source_class->get_metadata = mafw_fs_source_get_metadata;
	source_class->get_metadatas = mafw_fs_source_get_metadatas;

	g_type_class_add_private(klass, sizeof (MafwFsSourcePrivate));
}

static void
mafw_fs_source_init (MafwFsSource *source)
{
	source->priv = MAFW_FS_SOURCE_GET_PRIVATE (source);
	source->priv->browseid = 0;
}

/*__________________________ Utilities ____________________________ */

static GHashTable *
grab_metadata_from_file (GnomeVFSFileInfo *info, gchar *uri, gchar **keys)
{
	GHashTable *metadata = NULL;
	gchar *mime = NULL;
	gboolean is_dir = FALSE;
	gint i = 0;
	const gchar *const *real_keys = NULL;

	metadata = mafw_metadata_new ();

	/* There is a wildcard for requesting all supported keys */
	if (mafw_source_all_keys ((const gchar *const *) keys)) {
		real_keys = MAFW_SOURCE_LIST (MAFW_METADATA_KEY_TITLE,
					      MAFW_METADATA_KEY_URI,
					      MAFW_METADATA_KEY_MIME,
					      MAFW_METADATA_KEY_ADDED,
					      MAFW_METADATA_KEY_MODIFIED);
	} else {
		real_keys = (const gchar *const *) keys;
	}

	/* Extract some info from Gnome VFS, we could maybe
	   use tracker or some other tool to extract more
	   metadata */
	is_dir = info->type == GNOME_VFS_FILE_TYPE_DIRECTORY;
	mime = is_dir ? g_strdup (MAFW_METADATA_VALUE_MIME_CONTAINER) : 
		g_strdup (info->mime_type);

	/* Go through all requested keys, and fill the hashtable 
	   with those requested and supported */
	while (real_keys[i]) {
                /* We use the file name as title */
		if (!strcmp (real_keys[i], MAFW_METADATA_KEY_TITLE))
			mafw_metadata_add_str (metadata, MAFW_METADATA_KEY_TITLE,
					       g_strdup (info->name));
		
		/* Set URI */
		if (!strcmp (real_keys[i], MAFW_METADATA_KEY_URI))
			mafw_metadata_add_str (metadata, MAFW_METADATA_KEY_URI,
					       g_strdup (uri));
		
		/* Set ADDED and MODIFIED times */
		if (!strcmp (real_keys[i], MAFW_METADATA_KEY_ADDED))
			mafw_metadata_add_long (metadata, MAFW_METADATA_KEY_ADDED,
						info->mtime);
		if (!strcmp (real_keys[i], MAFW_METADATA_KEY_MODIFIED))
			mafw_metadata_add_long (metadata, 
						MAFW_METADATA_KEY_MODIFIED,
						info->ctime);
		
		/* Set mime-type, for container, we use a special mime type */
		if (!strcmp (real_keys[i], MAFW_METADATA_KEY_MIME))
			mafw_metadata_add_str (metadata, 
					       MAFW_METADATA_KEY_MIME,
					       mime);
		i++;
	}

	return metadata;
}

static gboolean
push_result (MafwSource *source,
	     guint browseid,
	     gint index,
	     gint pending,
	     gchar **keys,
	     GnomeVFSFileInfo *info, 
	     gchar *base_path,
	     MafwSourceBrowseResultCb browse_cb, 
	     gpointer user_data)

{
	g_return_val_if_fail (info, FALSE);

	GHashTable *metadata = NULL;
	gchar *objectid;
	gchar *uri;
	const gchar *home;
	
	/* Get metadata for this item */
	home = g_get_home_dir ();
	g_assert (home != NULL);
	uri = g_strconcat ("file://", home, "/", base_path, "/",
			   info->name, NULL);
	metadata = grab_metadata_from_file (info, uri, keys);

	/* Create the object identifier for this item */
	if (base_path[0] != '\0') {
		objectid = g_strconcat (MAFW_FS_SOURCE_UUID "::", 
					base_path, "/", info->name, NULL);
	}  else {
		objectid = g_strconcat (MAFW_FS_SOURCE_UUID "::",
					info->name, NULL);
	}

	/* Invoke the user callback with the information about this file */
	browse_cb (source,    /* The source object */
		   browseid,  /* The browse operation identifier */
		   pending,   /* # of pending results of this browse op. */
		   index,     /* Index of this item in the result set */
		   objectid,  /* The object identifier of this item */
		   metadata,  /* Item's metadata */
		   user_data, /* Callback user data */
		   NULL);     /* Error */

	/* Free resources */
	g_free (uri);
	g_free (objectid);
	mafw_metadata_release (metadata);

	return TRUE;
}

static void
push_empty_result (MafwSource *source,
		   guint browseid,
		   MafwSourceBrowseResultCb browse_cb, 
		   gpointer user_data)

{
	browse_cb (source,    /* The source object */
		   browseid,  /* The browse operation identifier */
		   0,         /* # of pending results of this browse op. */
		   -1,        /* Index of this item in the result set */
		   NULL,      /* The object identifier of this item */
		   NULL,      /* Item's metadata */
		   user_data, /* Callback user data */
		   NULL);     /* Error */
}

static GList *
filter_media_content (GList *files, guint offset, guint count)
{
	GList *iter;

	/* We are only interested in audio and video mimes (or directories) */
	iter = files;
	while (iter != NULL) {
		gboolean remove = FALSE;
		GnomeVFSFileInfo *info;
		GList *tmp;

		info = (GnomeVFSFileInfo *) iter->data;
		if ((!g_str_has_prefix (info->mime_type, "audio") &&
		     !g_str_has_prefix (info->mime_type, "video") &&
		     !g_str_has_prefix (info->mime_type, "x-directory")) ||
		    (!strcmp (info->name, "..") || !strcmp (info->name, "."))) {
			remove = TRUE;
		}
		
		tmp = iter;
		iter = g_list_next (iter);
		if (remove) {
			files = g_list_delete_link (files, tmp);
		}
	}

	/* Select the (offset, count) subset.
	   Count == 0 means all items from offset */
	while (offset--) {
		gnome_vfs_file_info_unref (files->data);
		files = g_list_delete_link (files, files);
	}
	iter = files;
	if (count > 0) {
		while (count--) {
			iter = g_list_next (iter);
		}
		while (iter != NULL) {
			GList *tmp = iter;
			iter = g_list_next (iter);
			gnome_vfs_file_info_unref (tmp->data);
			files = g_list_delete_link (files, tmp);
		}
	}
	
	return files;
}

static gboolean
browse_idle (gpointer user_data)
{
	gchar *path = NULL;
	gchar *abs_path = NULL;
	gchar *uri = NULL;
	const gchar *home ;
	GnomeVFSFileInfo *info = NULL;
	GnomeVFSResult res;
	GList *files = NULL;
	GList *iter = NULL;
	gint index;
	gint pending;
	struct browse_closure *c = NULL;
	
	c = (struct browse_closure *) user_data;

	/* Parse the object id (uuid::path), we want the path part 
	   We probably want to check that the UUID is correct too */
	mafw_source_split_objectid (c->objectid, NULL, &path);

	/* If this fails, then the objectid is not well formed */
	g_assert (path != NULL);

	/* Transform the path relative to $HOME into a valid URI */
	home = g_get_home_dir ();
	g_assert (home != NULL);
	abs_path = g_strconcat (home, "/", path, NULL);
	uri = g_strconcat ("file://", abs_path, NULL);

	/* We can only browse directories, so check file type */
	info = gnome_vfs_file_info_new ();
	res = gnome_vfs_get_file_info (uri, info, 
				       GNOME_VFS_FILE_INFO_GET_MIME_TYPE);
	g_assert (res == GNOME_VFS_OK);
	g_assert (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY);
	
	/* So it is a directory. Let's scan its contents */
	res = gnome_vfs_directory_list_load (&files,
					     abs_path,
					     GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
					     GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
	g_assert (res == GNOME_VFS_OK);

	/* Filter out video, audio and directories and 
	   apply offset and count */
	files = filter_media_content (files, c->offset, c->count);

	if (files == NULL) {
		push_empty_result (c->source, 
				   c->browseid, 
				   c->browse_cb, 
				   c->user_data);
	} else {
		index = 0;
		pending = g_list_length (files) - 1;
		iter = files;
		do {
			/* Probably we want to do this in the idle loop too */
			if (push_result (c->source, 
					 c->browseid,
					 index,
					 pending--,
					 c->keys,
					 iter->data, 
					 path, 
					 c->browse_cb, 
					 c->user_data)) {
				index++;
			}
		} while ((iter = g_list_next (iter)) != NULL);
	}

	/* Free the closure */
	g_free (c->objectid);
	g_strfreev (c->keys);
	g_object_unref (c->source);
	g_free (c);

	g_free (uri);
	g_free (path);
	g_free (abs_path);
	g_list_foreach (files, (GFunc) gnome_vfs_file_info_unref, NULL);
	g_list_free (files);

	return FALSE;
}

static gboolean
metadata_idle (gpointer user_data)
{
	gchar *path = NULL;
	const char *home = NULL;
	gchar *uri = NULL;
	GnomeVFSFileInfo *info = NULL;
	GHashTable *metadata = NULL;
	struct metadata_closure *c = NULL;
	GnomeVFSResult r;
	GError *error = NULL;

	c = (struct metadata_closure *) user_data;

	/* We need the file path, which we obtain from the objectid */
	mafw_source_split_objectid (c->objectid, NULL, &path);
	g_assert (path != NULL);
		
	/* Get info from the file */
	home = g_get_home_dir ();
	g_assert (home != NULL);
	uri = g_strconcat ("file://", home, "/", path, NULL);

	info = gnome_vfs_file_info_new ();
	r = gnome_vfs_get_file_info (uri,
				     info,
				     GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
				     GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
	
	if (r == GNOME_VFS_OK) { 
		/* Grab metadata for this file */
		metadata = grab_metadata_from_file (info, uri, c->keys);
	} else {
		error = g_error_new (MAFW_SOURCE_ERROR, 
				     MAFW_SOURCE_ERROR_INVALID_OBJECT_ID,
				     "Failed to open uri %s", uri);
	}

	/* Emit result */
	c->metadata_cb (c->source, c->objectid, metadata, c->user_data, error);

	/* Free resources */
	g_object_unref (c->source);
	g_free (c->objectid);
	g_strfreev (c->keys);
	g_free (c);
	if (error)
		g_error_free (error);

	g_free (path);
	g_free (uri);
	gnome_vfs_file_info_unref (info);
	mafw_metadata_release (metadata);

	return FALSE;
}

static gboolean
metadatas_idle (gpointer user_data)
{
	gint i = 0;
	GHashTable *metadata = NULL;
	const gchar *home = NULL;
	struct metadatas_closure *c = NULL;
	GnomeVFSResult r;

	c = (struct metadatas_closure *) user_data;

	/* Get home directory path */
	home = g_get_home_dir ();
	g_assert (home != NULL);

	/* Create the metadata hastable that we will fill */
	metadata = g_hash_table_new_full (g_str_hash,
					  g_str_equal,
					  (GDestroyNotify) g_free,
					  (GDestroyNotify) mafw_metadata_release);

	/* For each objectid, grab its metadata and emit */
	do {
		GHashTable *file_metadata = NULL;
		GnomeVFSFileInfo *info = NULL;
		gchar *uri = NULL;
		gchar *path = NULL;

		/* We need the file path, which we obtain from the objectid */
		mafw_source_split_objectid (c->objectids[i], NULL, &path);
		g_assert (path != NULL);

		/* Get info for this file */
		uri = g_strconcat ("file://", home, "/", path, NULL);

		info = gnome_vfs_file_info_new ();
		r = gnome_vfs_get_file_info (uri,
					     info,
					     GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
					     GNOME_VFS_FILE_INFO_FOLLOW_LINKS);

		/* Grab and store metadata for this file */
		if (r == GNOME_VFS_OK) {
			file_metadata =
				grab_metadata_from_file (info, uri, c->keys);
		}

		g_hash_table_insert (metadata, 
				     g_strdup (c->objectids[i]), 
				     file_metadata);

		/* Free resources */
		g_free (path);
		g_free (uri);
		gnome_vfs_file_info_unref (info);
	} while (c->objectids[++i]);

	/* Emit metadata to the client */
	c->metadata_cb (c->source, metadata, c->user_data, NULL);

	/* Free resources */
	g_object_unref (c->source);
	g_strfreev (c->objectids);
	g_strfreev (c->keys);
	g_free (c);

 	g_hash_table_unref (metadata);

	return FALSE;
}

/*_____________________ API implementation ________________________ */

static void
mafw_fs_source_get_metadatas (MafwSource *source,
			      const gchar **objectids,
			      const gchar *const *keys,
			      MafwSourceMetadataResultsCb metadata_cb,
			      gpointer user_data)
{
	struct metadatas_closure *closure = NULL;

	/* We need at least one objectid */
	g_assert (objectids != NULL);
	g_assert (objectids[0] != NULL);

	/* Execute op in idle loop */
	closure = g_new0 (struct metadatas_closure, 1);
	closure->source = g_object_ref (source);
	closure->objectids = g_strdupv ((gchar **) objectids);
	closure->keys = g_strdupv ((gchar **) keys);
	closure->metadata_cb = metadata_cb;
	closure->user_data = user_data;
	g_idle_add (metadatas_idle, closure);
}

static void
mafw_fs_source_get_metadata (MafwSource *source,
			     const gchar *objectid,
			     const gchar *const *keys,
			     MafwSourceMetadataResultCb metadata_cb,
			     gpointer user_data)
{
	struct metadata_closure *closure = NULL;

	/* Execute op in idle loop */
	closure = g_new0 (struct metadata_closure, 1);
	closure->source = g_object_ref (source);
	closure->objectid = g_strdup (objectid);
	closure->keys = g_strdupv ((gchar **) keys);
	closure->metadata_cb = metadata_cb;
	closure->user_data = user_data;
	g_idle_add (metadata_idle, closure);
}

static guint
mafw_fs_source_browse (MafwSource *source,
		       const gchar *objectid,
		       gboolean recursive,
		       const MafwFilter *filter,
		       const gchar *sorting,
		       const gchar *const *keys,
		       guint offset,
		       guint count,
		       MafwSourceBrowseResultCb browse_cb,
		       gpointer user_data)
{
	struct browse_closure *closure = NULL;;

	/* We need a valid filesystem source object */
	g_assert (MAFW_IS_FS_SOURCE (source));

	/* We must have an objectid representing the object to browse */
	g_assert (objectid != NULL);

	/* We must have a callback to provide the browse results */
	g_assert (browse_cb != NULL);

	/* We run the operation in the idle loop, since we want 
	   to return the browse identifier first to the client */

	/* We do not implement recursive browsing */
	/* We do not implement sorting */
	/* We do not implement filtering */

	closure = g_new0 (struct browse_closure, 1);
	closure->source = g_object_ref (source);
	closure->objectid = g_strdup (objectid);
	closure->keys = g_strdupv ((gchar **) keys);
	closure->offset = offset;
	closure->count = count;
	closure->browse_cb = browse_cb;
	closure->user_data = user_data;
	closure->browseid = MAFW_FS_SOURCE (source)->priv->browseid;
	g_idle_add (browse_idle, closure);

	/* Return the identifier of this browse operation */
	return MAFW_FS_SOURCE (source)->priv->browseid++;
}
