#include <glib.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#include "../src/mafw-fs-source.h"

static MafwSource *source = NULL;
static GMainLoop *loop = NULL;

static void
print_metadata (GHashTable *metadata, gchar **keys)
{
	GValue *value;
	const gchar *title, *mime, *uri;
	glong date;
	gchar *date_str;
	gint i=0;

	while (keys[i])  {
		if (!strcmp (keys[i], MAFW_METADATA_KEY_TITLE)) {
			value = mafw_metadata_first (metadata,
						     MAFW_METADATA_KEY_TITLE);
			if (value == NULL) {
				title = "Unknown";
			} else {
				title = g_value_get_string (value);
			}
			g_print ("\tTitle: %s\n", title);
		}

		if (!strcmp (keys[i], MAFW_METADATA_KEY_URI)) {
			value = mafw_metadata_first (metadata,
						     MAFW_METADATA_KEY_URI);
			if (value == NULL) {
				uri = "Unknown";
			} else {
				uri = g_value_get_string (value);
			}
			g_print ("\tURI: %s\n", uri);
		}

		if (!strcmp (keys[i], MAFW_METADATA_KEY_MIME)) {
			value = mafw_metadata_first (metadata,
						     MAFW_METADATA_KEY_MIME);
			if (value == NULL) {
				mime = "Unknown";
			} else {
				mime = g_value_get_string (value);
			}
			g_print ("\tMime Type: %s\n", mime);
		}

		if (!strcmp (keys[i], MAFW_METADATA_KEY_ADDED)) {
			value = mafw_metadata_first (metadata, MAFW_METADATA_KEY_ADDED);
			if (value == NULL) {
				date_str = "Unknown\n";
			} else {
				date = g_value_get_long (value);
				date_str = ctime (&date);
			}
			g_print ("\tAdded Time: %s", date_str);
		}

		if (!strcmp (keys[i], MAFW_METADATA_KEY_MODIFIED)) {
			value = mafw_metadata_first (metadata, MAFW_METADATA_KEY_MODIFIED);
			if (value == NULL) {
				date_str = "Unknown\n";
			} else {
				date = g_value_get_long (value);
				date_str = ctime (&date);
			}
			g_print ("\tModification Time: %s", date_str);
		}
		i++;
	}
}

static void
init_test (void)
{
	GError *error = NULL;
	MafwRegistry *registry = NULL;
	GList *sources;

	g_type_init ();

	registry = MAFW_REGISTRY (mafw_registry_get_instance ());
	sources = mafw_registry_get_sources (registry);
	if (sources == NULL) {
		mafw_fs_source_plugin_initialize (registry, &error);
		sources = mafw_registry_get_sources (registry);
	}

	if (sources == NULL) {
		g_critical ("Failed to load plugin");
		if (error != NULL) {
			g_critical ("%s", error->message);
		}
		g_assert_not_reached ();
	}

	source = MAFW_SOURCE (sources->data);
	
	g_assert (source != NULL);
}

static void
browse_cb (MafwSource *source, 
	   guint browseid,
	   gint pending,
	   guint index,
	   const gchar *objectid,
	   GHashTable *metadata,
	   gpointer user_data,
	   const GError *error)
{
	if (error != NULL) {
		g_error ("Error: %s", error->message);
	}

	if (index == 0) {
		g_print ("Got results (%d) for browse operation %u\n", 
			 pending + 1, browseid);
	}

	g_print ("%d objectid: %s\n", index, objectid);

	if (metadata == NULL) {
		g_print ("\t Could not retrieve metadata");
		return;
	}
	
	print_metadata (metadata, (gchar **) user_data);
	g_main_loop_quit (loop);
}

static void
metadata_cb (MafwSource *source,
	     const gchar *objectid,
	     GHashTable *metadata,
	     gpointer user_data,
	     const GError *error)
{
	if (error != NULL) {
		g_error ("Error: %s", error->message);
	}
	
	if (metadata == NULL) {
		g_print ("\t Could not retrieve metadata for %s", objectid);
		return;
	}

	g_print ("Got metadata for %s\n", objectid);
	
	print_metadata (metadata, (gchar **) user_data);
	g_main_loop_quit (loop);
}

static void
metadatas_cb (MafwSource *source,
	      GHashTable *metadatas,
	      gpointer user_data,
	      const GError *error)
{
	GList *objectids = NULL;

	if (error != NULL) {
		g_error ("Error: %s", error->message);
	}
	
	objectids = (GList *) g_hash_table_get_keys (metadatas);
	while (objectids != NULL) {
		gchar *objectid = (gchar *) objectids->data;
		GHashTable *m = g_hash_table_lookup (metadatas, objectid);
	        if (m == NULL) {
	                g_print ("Could not retrieve metadata for %s\n", objectid);
        	} else {
		        g_print ("Got metadata for %s\n", objectid);
		        print_metadata (m, (gchar **) user_data);
		}
		objectids = g_list_next (objectids);
	}
	g_main_loop_quit (loop);
}

static void
test_browse (gchar *objectid)
{
	const gchar *const *keys;

 	keys = MAFW_SOURCE_LIST (MAFW_METADATA_KEY_TITLE,
				 MAFW_METADATA_KEY_URI,
				 MAFW_METADATA_KEY_MIME,
				 MAFW_METADATA_KEY_ADDED,
				 MAFW_METADATA_KEY_MODIFIED);

	mafw_source_browse (source, 
			    objectid,
			    FALSE,
			    NULL,
			    NULL,
			    keys,
			    10,
			    15, 
			    browse_cb,
			    (gpointer) g_strdupv ( (gchar **) keys));
}

static void
test_metadata (gchar *objectid)
{
	const gchar *const *keys;

 	keys = MAFW_SOURCE_LIST (MAFW_METADATA_KEY_TITLE,
				 MAFW_METADATA_KEY_URI,
				 MAFW_METADATA_KEY_MIME,
				 MAFW_METADATA_KEY_ADDED,
				 MAFW_METADATA_KEY_MODIFIED);

	mafw_source_get_metadata (source, objectid, keys, metadata_cb, 
				  (gpointer) g_strdupv ((gchar **) keys));
}

static void
test_metadatas (const gchar **objectids)
{
	const gchar *const *keys;

 	keys = MAFW_SOURCE_LIST (MAFW_METADATA_KEY_TITLE,
				 MAFW_METADATA_KEY_URI,
				 MAFW_METADATA_KEY_MIME,
				 MAFW_METADATA_KEY_ADDED,
				 MAFW_METADATA_KEY_MODIFIED);

	mafw_source_get_metadatas (source, objectids, keys, metadatas_cb,
				   (gpointer) g_strdupv ((gchar **) keys));
}

int
main (gint argc, gchar *argv[])
{
	if ((argc > 3 && strcmp (argv[1], "metadatas")) || argc < 3) {
		g_print ("test <browse|metadata|metadatas> <objectid>\n");
		exit (1);
	}

	init_test ();

	if (!strcmp (argv[1], "browse")) {
		test_browse (argv[2]);
	} else if (!strcmp (argv[1], "metadata")) {
		test_metadata (argv[2]);
	} else if (!strcmp (argv[1], "metadatas")) {
		test_metadatas ((const gchar **) &argv[2]);
	} else {
		g_print ("test <browse|metadata|metadatas> <objectid>\n");
		exit (1);
	}

	loop = g_main_loop_new (NULL, FALSE);
	g_main_loop_run (loop);

	return 0;
}
